@@ -13,11 +13,11 @@ XPath 2.0 implementation - part 2 (operators, expressions and multi-role tokens)
 import math
 import operator
 from copy import copy
+from collections.abc import Iterator
 from decimal import Decimal, DivisionByZero
-from typing import cast, List, Type, Union
+from typing import cast, Union
 
-from elementpath._typing import Iterator
-from elementpath.aliases import Emptiable
+from elementpath.aliases import Emptiable, SequenceType
 from elementpath.protocols import XsdAttributeProtocol
 from elementpath.exceptions import ElementPathError
 from elementpath.helpers import OCCURRENCE_INDICATORS, numeric_equal, numeric_not_equal, \
@@ -25,8 +25,8 @@ from elementpath.helpers import OCCURRENCE_INDICATORS, numeric_equal, numeric_no
 from elementpath.namespaces import XSD_NAMESPACE, XSD_NOTATION, XSD_ANY_ATOMIC_TYPE, \
     XSD_UNTYPED, get_namespace, get_expanded_name
 from elementpath.datatypes import UntypedAtomic, QName, AnyURI, \
-    Duration, Integer, DoubleProxy10, AnyAtomicType, AtomicType, NumericType
-from elementpath.decoder import get_atomic_value
+    Duration, Integer, DoubleProxy10, AtomicType, NumericType
+from elementpath.decoder import get_atomic_sequence
 from elementpath.xpath_nodes import ElementNode, DocumentNode, XPathNode, AttributeNode
 from elementpath.sequence_types import is_instance
 from elementpath.xpath_context import ContextType, ItemType, XPathSchemaContext
@@ -66,7 +66,7 @@ def nud_variable_reference(self: XPathToken) -> XPathToken:
 
 @method('$')
 def evaluate_variable_reference(self: XPathToken, context: ContextType = None) \
-        -> Emptiable[Union[ItemType, List[ItemType]]]:
+        -> Emptiable[SequenceType[ItemType]]:
     if context is None:
         raise self.missing_context()
 
@@ -102,7 +102,11 @@ def evaluate_variable_reference(self: XPathToken, context: ContextType = None) \
                 else:
                     if self.parser.schema is not None:
                         xsd_type = self.parser.schema.get_type(type_name)
-                        return cast(AnyAtomicType, get_atomic_value(xsd_type))
+                        result = [v for v in get_atomic_sequence(xsd_type)]
+                        if len(result) == 1:
+                            return result[0]
+                        else:
+                            return cast(Emptiable[AtomicType], result)
 
             return UntypedAtomic('1')
 
@@ -127,9 +131,9 @@ def select_intersect_and_except_operators(self: XPathToken, context: ContextType
         raise self.error('XPTY0004', 'only XPath nodes are allowed')
 
     if self.symbol == 'except':
-        yield from cast(List[XPathNode], sorted(s1 - s2, key=node_position))
+        yield from cast(list[XPathNode], sorted(s1 - s2, key=node_position))
     else:
-        yield from cast(List[XPathNode], sorted(s1 & s2, key=node_position))
+        yield from cast(list[XPathNode], sorted(s1 & s2, key=node_position))
 
 
 ###
@@ -151,7 +155,7 @@ def nud_if_expression(self: XPathToken) -> XPathToken:
 
 @method('if')
 def evaluate_if_expression(self: XPathToken, context: ContextType = None) \
-        -> Union[ItemType, List[ItemType]]:
+        -> Union[ItemType, list[ItemType]]:
     if self.boolean_value(self[0].select(copy(context))):
         if isinstance(context, XPathSchemaContext):
             self[2].evaluate(copy(context))
@@ -292,7 +296,6 @@ def evaluate_instance_expression(self: XPathToken, context: ContextType = None)
             raise self.missing_context()
 
         context = copy(context)
-
         for position, context.item in enumerate(self[0].select(context)):
             if context.axis is None:
                 context.axis = 'self'
@@ -327,7 +330,7 @@ def evaluate_instance_expression(self: XPathToken, context: ContextType = None)
 
 @method('treat')
 def evaluate_treat_expression(self: XPathToken, context: ContextType = None) \
-        -> List[ItemType]:
+        -> list[ItemType]:
     occurs = self[1].occurrence
     position = None
     castable_expr = []
@@ -431,7 +434,7 @@ def evaluate_cast_expressions(self: XPathToken, context: ContextType = None) \
         else:
             local_name = atomic_type.split('}')[1]
             try:
-                token_class = cast(Type[XPathConstructor], self.parser.symbol_table[local_name])
+                token_class = cast(type[XPathConstructor], self.parser.symbol_table[local_name])
             except KeyError:
                 msg = f"atomic type {type_name!r} not found in the in-scope schema types"
                 raise self.error('XPST0051', msg)
@@ -471,7 +474,7 @@ def evaluate_cast_expressions(self: XPathToken, context: ContextType = None) \
 # Comma operator - concatenate items or sequences
 @method(infix(',', bp=5))
 def evaluate_comma_operator(self: XPathToken, context: ContextType = None) \
-        -> List[ItemType]:
+        -> list[ItemType]:
     results = []
     for op in self:
         result = op.evaluate(context)
@@ -523,7 +526,7 @@ def led_parenthesized_expression(self: XPathToken, left: XPathToken) -> XPathTok
 
 @method('(')
 def evaluate_parenthesized_expression(self: XPathToken, context: ContextType = None) \
-        -> Union[ItemType, List[ItemType]]:
+        -> Union[ItemType, list[ItemType]]:
     return self[0].evaluate(context) if self else []
 
 
@@ -569,10 +572,10 @@ def evaluate_value_comparison_operators(self: XPathToken, context: ContextType =
     elif all(isinstance(x, DoubleProxy10) for x in operands):
         # Special case of two <class 'float'> values: use custom operators
         if self.symbol == 'eq':
-            return numeric_equal(*cast(List[float], operands))
+            return numeric_equal(*cast(list[float], operands))
         elif self.symbol == 'ne':
-            return numeric_not_equal(*cast(List[float], operands))
-        elif numeric_equal(*cast(List[float], operands)):
+            return numeric_not_equal(*cast(list[float], operands))
+        elif numeric_equal(*cast(list[float], operands)):
             return self.symbol in ('le', 'ge')
 
     cls0, cls1 = type(operands[0]), type(operands[1])
@@ -667,7 +670,7 @@ def led_range_expression(self: XPathToken, left: XPathToken) -> XPathToken:
 
 
 @method('to')
-def evaluate_range_expression(self: XPathToken, context: ContextType = None) -> List[int]:
+def evaluate_range_expression(self: XPathToken, context: ContextType = None) -> list[int]:
     start, stop = self.get_operands(context, cls=Integer)
     try:
         return [x for x in range(start, stop + 1)]
@@ -677,7 +680,7 @@ def evaluate_range_expression(self: XPathToken, context: ContextType = None) ->
 
 @method('to')
 def select_range_expression(self: XPathToken, context: ContextType = None) -> Iterator[int]:
-    yield from cast(List[int], self.evaluate(context))
+    yield from cast(list[int], self.evaluate(context))
 
 
 ###
@@ -779,20 +782,16 @@ def select_element_kind_test(self: XPathFunction, context: ContextType = None) \
             if len(self) == 1:
                 yield cast(ElementNode, item)  # Already selected by sequence type test
             elif isinstance(item, ElementNode):
-                try:
-                    type_annotation = get_expanded_name(self[1].source, self.parser.namespaces)
-                except KeyError:
-                    type_annotation = self[1].source
-
+                type_annotation = self[1].name
                 if item.nilled:
-                    if type_annotation[-1] in '*?':
+                    if self[1].occurrence in ('*', '?'):
                         yield item
-                elif item.xsd_type is not None:
-                    if type_annotation == item.xsd_type.name:
+                elif item.type_name == type_annotation:
+                    if type_annotation != XSD_UNTYPED:
                         yield item
-                    elif is_instance(item.typed_value, type_annotation, self.parser):
+                    elif self[0].symbol != '*':
                         yield item
-                elif type_annotation == XSD_UNTYPED and self[0].symbol != '*':
+                elif is_instance(item.typed_value, type_annotation, self.parser):
                     yield item
 
 
@@ -849,11 +848,10 @@ def select_schema_element_kind_test(self: XPathFunction, context: ContextType =
 
     if self.parser.schema is not None:
         for _ in context.iter_children_or_self():
-            if self.parser.schema.get_element(qname) is None \
-                    and self.parser.schema.get_substitution_group(qname) is None:
+            if self.parser.schema.get_element(qname) is None:
                 raise self.error('XPST0008', "element %r not found in schema" % element_name)
 
-            if isinstance(context.item, ElementNode) and context.item.elem.tag == qname:
+            if isinstance(context.item, ElementNode) and context.item.name == qname:
                 yield context.item
                 return
 
@@ -927,7 +925,7 @@ def select_attribute_kind_test_or_axis(self: XPathToken, context: ContextType =
             yield from cast(Iterator[AttributeNode], self[0].select(context))
     elif not self:
         for attribute in context.iter_attributes():
-            yield attribute.value
+            yield attribute
     else:
         name = self[0].value
         assert isinstance(name, str)
@@ -941,24 +939,14 @@ def select_attribute_kind_test_or_axis(self: XPathToken, context: ContextType =
         for attribute in context.iter_attributes():
             if attribute.match_name(name):
                 if isinstance(context, XPathSchemaContext):
-                    self.add_xsd_type(attribute)
                     continue
 
-                if attribute.xsd_type is None:
-                    attribute.xsd_type = self.get_xsd_type(attribute)
-
-                if type_name and attribute.xsd_type is None:
-                    if type_name == XSD_UNTYPED and name != '*':
-                        yield attribute.value
-                elif not type_name or attribute.xsd_type is not None and \
-                        attribute.xsd_type.name == type_name or \
+                if type_name == XSD_UNTYPED == attribute.type_name:
+                    if name != '*':
+                        yield attribute
+                elif not type_name or attribute.type_name == type_name or \
                         is_instance(attribute.typed_value, type_name, self.parser):
-                    typed_value = attribute.typed_value
-                    if typed_value is not None:
-                        if isinstance(typed_value, list):
-                            yield from typed_value
-                        else:
-                            yield typed_value
+                    yield attribute
 
 
 # XPath 2.0 definitions continue into module xpath2_functions
